// HTMLViewerView.cpp : implementation of the CHTMLViewerView class
//

#include "stdafx.h"
#include "HTMLViewer.h"

#include "HTMLViewerDoc.h"
#include "HTMLViewerView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//#define APP ((CHTMLViewerApp* )AfxGetApp())

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView

IMPLEMENT_DYNCREATE(CHTMLViewerView, CScrollView)

BEGIN_MESSAGE_MAP(CHTMLViewerView, CScrollView)
	//{{AFX_MSG_MAP(CHTMLViewerView)
	ON_WM_LBUTTONDOWN()
	ON_WM_CHAR()
	ON_WM_KEYDOWN()
	ON_WM_SETFOCUS()
	ON_WM_KILLFOCUS()
	ON_WM_SIZE()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CScrollView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView construction/destruction

CHTMLViewerView::CHTMLViewerView() : m_rcMargins(20, 20, 20, 20)
{	
	m_currPos.charPos.nIndex = 0;
	m_currPos.charPos.pTokenInfo = NULL;
	m_currPos.pLine = NULL;
	m_currPos.bIsLineEnd = false;
	
	m_caretPos.x = 20;
	m_caretPos.y = 20;
	m_nCaretHeight = 15;	

	m_bSelection = false;
}

CHTMLViewerView::~CHTMLViewerView()
{
}

BOOL CHTMLViewerView::PreCreateWindow(CREATESTRUCT& cs)
{
	// The window class should be modified to set
	// hCursor whit NULL value. Its needed to change
	// mouse curso to hand cursor when its over a hyperlink.
	// In that way it won't be necesary to capture cursor.
	WNDCLASS wc;

	// Start setting with common view values
	wc.style = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = AfxWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = AfxGetInstanceHandle();
    wc.hCursor = NULL;
	wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
	wc.lpszMenuName = NULL;
    wc.lpszClassName = "HTMLViewClass"; 

	// hIcon needs a special code that I borrowed from MFC code
	HINSTANCE hInst = AfxFindResourceHandle(
		MAKEINTRESOURCE(AFX_IDI_STD_FRAME), RT_GROUP_ICON);
	if ((wc.hIcon = ::LoadIcon(hInst, MAKEINTRESOURCE(AFX_IDI_STD_FRAME))) == NULL)
	{
		// use default icon
		wc.hIcon = ::LoadIcon(NULL, IDI_APPLICATION);
	}

	// Register window class
	if (!AfxRegisterClass(&wc)) {
		TRACE0("Failed to register view window class\n");
		return 0;
	}

	// Say MFC to use the just created class
	cs.lpszClass = wc.lpszClassName;

	return CScrollView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView drawing

void CHTMLViewerView::OnDraw(CDC* pDC)
{
	//TestDrawing(pDC);
	
	DrawLine(pDC);
}

#ifdef _DEBUG
void CHTMLViewerView::TestDrawing(CDC* pDC)
{
	POSITION pos = m_page.GetHeadPosition();
	CString str;
	TEXTLINE line;
	int x = 20, y = 20;
	CTokenInfo first, last;
	TOKENLIST& tiList = GetTokenList();

	while(pos)
	{
		line = m_page.GetNext(pos);
		first = tiList.GetAt(line.first.pTokenInfo);
		last = tiList.GetAt(line.last.pTokenInfo);

		str.Format("FirstTK:%d, FistrChar:%d, TKLen:%d, LastTK:%d, LastChar:%d, TKLen:%d", 
			line.first.pTokenInfo, line.first.nIndex, 
			first.m_nLen,	
			line.last.pTokenInfo, line.last.nIndex,
			last.m_nLen
			);

		pDC->TextOut(x, y, str);
		y += 20;
		pDC->TextOut(x, y, first.m_pText, first.m_nLen);
		y += 20;
		pDC->TextOut(x, y, last.m_pText, last.m_nLen);
		y += 20;
	}
}
#endif

void CHTMLViewerView::DrawLine(CDC* pDC)
{
	TOKENLIST& tiList = GetTokenList();
	POSITION pos = tiList.GetHeadPosition();
	CTokenInfo ti;
	CFont *pOldFont, font;
	CRect rect;
	GetClientRect(rect);
	rect.DeflateRect(m_rcMargins);

	int x = rect.left, y = rect.top;

	POSITION linePos = m_page.GetHeadPosition();
	POSITION tiPos;
	TEXTLINE line;
	
	pOldFont = pDC->GetCurrentFont();

	pDC->SetTextAlign(TA_UPDATECP);

	while (linePos) 
	{
		pDC->MoveTo(x, y);

		line = m_page.GetNext(linePos);

		tiPos = line.first.pTokenInfo;
		ASSERT(tiPos);
		
		// Get first token
		ti = tiList.GetNext(tiPos);
		
		PrepareFont(&font, ti.m_nStyle);
		pDC->SelectObject(&font);

		if (ti.m_nStyle & TI_HYPERLINK)
			pDC->SetTextColor(RGB(0,0,255));
		else
			pDC->SetTextColor(RGB(0,0,0));

		// Set text pointer and text lenght
		ti.m_pText += line.first.nIndex;
		
		if (line.first.pTokenInfo == line.last.pTokenInfo)
			ti.m_nLen = line.last.nIndex - line.first.nIndex + 1;
		else
			ti.m_nLen -= line.first.nIndex;

		pDC->TextOut(x, y, ti.m_pText, ti.m_nLen);

		pDC->SelectObject(pOldFont);
		font.DeleteObject();

		// End first token
		
		if (line.first.pTokenInfo != line.last.pTokenInfo)
		{
			// Get intermediate tokens
			while (tiPos != line.last.pTokenInfo)
			{
				ASSERT(tiPos);
				ti = tiList.GetNext(tiPos);

				PrepareFont(&font, ti.m_nStyle);
				pDC->SelectObject(&font);

				if (ti.m_nStyle & TI_HYPERLINK)
					pDC->SetTextColor(RGB(0,0,255));
				else
					pDC->SetTextColor(RGB(0,0,0));
				
				pDC->TextOut(x, y, ti.m_pText, ti.m_nLen);

				pDC->SelectObject(pOldFont);
				font.DeleteObject();
			} 
			// End intermediate tokens

			// Get last token
			ti = tiList.GetAt(tiPos);

			PrepareFont(&font, ti.m_nStyle);
			pDC->SelectObject(&font);

			if (ti.m_nStyle & TI_HYPERLINK)
				pDC->SetTextColor(RGB(0,0,255));
			else
				pDC->SetTextColor(RGB(0,0,0));
			
			pDC->TextOut(x, y, ti.m_pText, line.last.nIndex + 1);

			pDC->SelectObject(pOldFont);
			font.DeleteObject();
			// End last token
		}

		y += line.nHeight;
		x = rect.left;
	}
}

/*void CHTMLViewerView::DrawSelection(CDC* pDC)
{
	if (!m_bSelection) return;

	LINETEXT line;
	POSITION pos = m_selStart.m_pLine;

	do
	{
		line = m_page.GetNext(pos);

	} while (pos != m_selEnd.m_pLine);
}*/

void CHTMLViewerView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();
	CSize sizeTotal;
	sizeTotal.cx = sizeTotal.cy = 100;
	SetScrollSizes(MM_TEXT, sizeTotal);
}

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView printing

BOOL CHTMLViewerView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CHTMLViewerView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

void CHTMLViewerView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView diagnostics

#ifdef _DEBUG
void CHTMLViewerView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CHTMLViewerView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}

CHTMLViewerDoc* CHTMLViewerView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CHTMLViewerDoc)));
	return (CHTMLViewerDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CHTMLViewerView message handlers

void CHTMLViewerView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	RELCHARPOS rcp;
	
	CPoint pt = GetCharFromPt(point, rcp);
	MoveCaretTo(pt, rcp);

	/*CString str;

	str.Format("Char:%c", 
		GetTokenList().GetAt(rcp.charPos.pTokenInfo).m_pText[rcp.charPos.nIndex]);

	MessageBox(str);*/

	/*if (cp.pTokenInfo != NULL)
	{
		MoveCaretTo(cp);
	}*/
	
	CScrollView::OnLButtonDown(nFlags, point);
}

int CHTMLViewerView::CalcLineBreak(int* pWidths, int nLen, int cx, int& x)
{
	ASSERT(pWidths && nLen > 0 && cx >= 0);
	int i = 0, dx;
	x = 0;

	do {
		dx = pWidths[i];
		if (x + dx <= cx) {
			x += dx;
			i++;
		}
		else return i;
	} while(i < nLen);
	
	return nLen;
}

void CHTMLViewerView::PrepareFont(CFont* pFont, int nStyle)
{
	if (nStyle & TI_USERFONT) {
	}
	else if (nStyle & TI_H1) {
		pFont->CreateFont(-20, 0, 0, 0, 
			700, 
			(nStyle & TI_ITALIC) ? TRUE:FALSE, 
			(nStyle & TI_UNDERLINE || nStyle & TI_HYPERLINK) ? TRUE:FALSE, 
			0, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
			DEFAULT_PITCH | FF_SWISS, "Arial");
	}
	else {
		// TI_TEXT
		pFont->CreateFont(-16, 0, 0, 0, 
			(nStyle & TI_BOLD) ? 700:400, 
			(nStyle & TI_ITALIC) ? TRUE:FALSE, 
			(nStyle & TI_UNDERLINE || nStyle & TI_HYPERLINK) ? TRUE:FALSE, 
			0, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
			DEFAULT_PITCH | FF_ROMAN, "Times New Roman");
	}
}

void CHTMLViewerView::PreparePage()
{
	TOKENLIST& list = GetTokenList();;
	POSITION pos = list.GetHeadPosition();
	CDC* pDC = GetDC();
	CFont* pOldFont = pDC->GetCurrentFont();
	CTokenInfo* pTI;
	CFont font;
	int i;

	// Walk by the hole token list sizing the char widths
	while (pos)
	{
		pTI = &list.GetNext(pos);

		// Make the font based on the style
		PrepareFont(&font, pTI->m_nStyle);

		// Select the font in the DC
		pDC->SelectObject(&font);

		// Alloc memory if its the first sizing
		if (pTI->m_pWidths == NULL)
			pTI->m_pWidths = new int[pTI->m_nLen];
		
		pTI->m_nHeight = pDC->GetTextExtent(&(pTI->m_pText[0]), 1).cy;
		
		// Size char widths
		for (i = 0; i < pTI->m_nLen; i++)
			pTI->m_pWidths[i] = pDC->GetTextExtent(&(pTI->m_pText[i]), 1).cx;

		// Restore previous font and delete current font
		pDC->SelectObject(pOldFont);
		font.DeleteObject();
	}

	ReleaseDC(pDC);
}

void CHTMLViewerView::BreakLines()
{
	CTokenInfo ti;
	TOKENLIST& list = GetTokenList();
	POSITION nextPos = list.GetHeadPosition();
	POSITION currPos, prevPos;
	TEXTLINE line;
	int nRemLen;	// Remnant Lenght 
	int nPrevLen;	// Previous Lenght
	int x, dx, nCharCount;
	CRect rect;
	
	GetClientRect(rect);
	rect.DeflateRect(m_rcMargins);
	
	m_page.RemoveAll();
	x = rect.left;
	line.first.pTokenInfo = nextPos;
	line.first.nIndex = 0;
	line.nHeight = 0;

	currPos = NULL;

	while (nextPos) 
	{
		prevPos = currPos;
		currPos = nextPos;
		ti = list.GetNext(nextPos);

		// Is new liene required
		if ((ti.m_nStyle & TI_NEWLINE) && prevPos != NULL)
		{
			line.last.pTokenInfo = prevPos;
			line.last.nIndex = nPrevLen - 1;
			m_page.AddTail(line);
			line.first.pTokenInfo = currPos;
			line.first.nIndex = 0;
			line.nHeight = 0;
			x = rect.left;
		}
		
		if (line.nHeight < ti.m_nHeight)
			line.nHeight = ti.m_nHeight;

		for (nRemLen = ti.m_nLen; nRemLen > 0; nRemLen -= nCharCount)
		{
			nCharCount = CalcLineBreak(ti.m_pWidths, nRemLen, 
				rect.right - x, dx);

			if (nCharCount < nRemLen) 
			{
				// Avoid a possible endesly loop
				if (nCharCount == 0 && x == rect.left) {
					TRACE("Too small window!!!");
					return;
				}
				
				// Is new line and new token. Special case
				if (nRemLen == ti.m_nLen && nCharCount == 0)
				{
					line.last.pTokenInfo = prevPos;
					line.last.nIndex = nPrevLen - 1;	// The whole previous string
					m_page.AddTail(line);
					line.first.pTokenInfo = currPos;
					line.first.nIndex = 0;
					line.nHeight = ti.m_nHeight;
				}
				else
				{
					line.last.pTokenInfo = currPos;
					line.last.nIndex = ti.m_nLen - (nRemLen - nCharCount) - 1;
					m_page.AddTail(line);
					line.first.pTokenInfo = currPos;
					line.first.nIndex = line.last.nIndex + 1;
					line.nHeight = ti.m_nHeight;
				}

				ti.m_pWidths += nCharCount;
				
				// Reset line coordinate
				x = rect.left;
			}
			else
			{
				x += dx;
			}
		}
		// I may need this lenght later
		nPrevLen = ti.m_nLen;
	}

	// There is still one more line to add. The page last line
	if (currPos != NULL)
	{
		line.last.pTokenInfo = currPos;
		line.last.nIndex = nPrevLen - 1;
		m_page.AddTail(line);
	}
}

void CHTMLViewerView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	//cstring str;
	//str.format("right = %d, char = %d, flags = %d", vk_right, nchar, nflags);
	//if (nchar == vk_right)
	//	messagebox(str);
	CScrollView::OnChar(nChar, nRepCnt, nFlags);
}

void CHTMLViewerView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	PreparePage();
	BreakLines();

	// Reset current line and char position
	m_currPos.pLine = m_page.GetHeadPosition();
	m_currPos.charPos.pTokenInfo = GetTokenList().GetHeadPosition();
	m_currPos.charPos.nIndex = 0;

	// Draw new data
	Invalidate();
}

void CHTMLViewerView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (m_page.IsEmpty())
		return;

	switch (nChar) {
		case VK_LEFT: 
		{
			if (m_currPos.bIsLineEnd) m_currPos.bIsLineEnd = false;
			TEXTLINE line = m_page.GetAt(m_currPos.pLine);
			if (!IsLineFirstChar(line, m_currPos.charPos))
			{
				m_caretPos.x -= GetPrevWidth(&m_currPos.charPos);
			}
			else if (m_currPos.pLine != m_page.GetHeadPosition()) 
			{
				m_page.GetPrev(m_currPos.pLine);
				line = m_page.GetAt(m_currPos.pLine);

				m_caretPos.y -= line.nHeight;
				m_caretPos.x = m_rcMargins.left;

				m_currPos.charPos.pTokenInfo = line.first.pTokenInfo;
				m_currPos.charPos.nIndex = line.first.nIndex;

				while (!IsLineLastChar(line, m_currPos.charPos))
					m_caretPos.x += GetNextWidth(&m_currPos.charPos);
			}
			break;
		}
		case VK_RIGHT:
		{
			TEXTLINE line = m_page.GetAt(m_currPos.pLine);
			if (!IsLineLastChar(line, m_currPos.charPos) && !m_currPos.bIsLineEnd)
			{
				m_caretPos.x += GetNextWidth(&m_currPos.charPos);
			}
			else if (!m_currPos.bIsLineEnd)
			{
				m_caretPos.x += GetNextWidth(&m_currPos.charPos);
				m_currPos.bIsLineEnd = true;
			}
			else if (m_currPos.pLine != m_page.GetTailPosition()) 
			{
				m_caretPos.y += line.nHeight;
				m_caretPos.x = m_rcMargins.left;

				m_page.GetNext(m_currPos.pLine);
				line = m_page.GetAt(m_currPos.pLine);

				m_currPos.charPos.pTokenInfo = line.first.pTokenInfo;
				m_currPos.charPos.nIndex = line.first.nIndex;
				m_currPos.bIsLineEnd = false;
			}
			break;
		}
		case VK_UP:
			if (m_currPos.pLine != m_page.GetHeadPosition()) 
			{
				m_page.GetPrev(m_currPos.pLine);
				TEXTLINE line = m_page.GetAt(m_currPos.pLine);

				m_caretPos.y -= line.nHeight;

				m_currPos.charPos.pTokenInfo = line.first.pTokenInfo;
				m_currPos.charPos.nIndex = line.first.nIndex;

				int x = m_rcMargins.left;

				while (x < m_caretPos.x && 
					!IsLineLastChar(line, m_currPos.charPos))
				{
					x += GetNextWidth(&m_currPos.charPos);
				}
				
				m_caretPos.x = x;
				if (m_currPos.bIsLineEnd) m_currPos.bIsLineEnd = false;
			}
			break;
		case VK_DOWN:
			if (m_currPos.pLine != m_page.GetTailPosition()) 
			{
				TEXTLINE line = m_page.GetNext(m_currPos.pLine);
				
				m_caretPos.y += line.nHeight;	// Current line height

				line = m_page.GetAt(m_currPos.pLine);	// Get next line data

				m_currPos.charPos.pTokenInfo = line.first.pTokenInfo;
				m_currPos.charPos.nIndex = line.first.nIndex;

				int x = m_rcMargins.left;

				while (x < m_caretPos.x && 
					!IsLineLastChar(line, m_currPos.charPos))
				{
					x += GetNextWidth(&m_currPos.charPos);
				}
				
				if (x > m_caretPos.x)
					x -= GetPrevWidth(&m_currPos.charPos);
				
				m_caretPos.x = x;
				if (m_currPos.bIsLineEnd) m_currPos.bIsLineEnd = false;
			}
			break;
		case VK_HOME:
		{
			if (m_currPos.bIsLineEnd) m_currPos.bIsLineEnd = false;
			TEXTLINE line = m_page.GetAt(m_currPos.pLine);
			m_currPos.charPos.pTokenInfo = line.first.pTokenInfo;
			m_currPos.charPos.nIndex = line.first.nIndex;
			m_caretPos.x = m_rcMargins.left;
			break;
		}
		case VK_END:
		{
			m_currPos.bIsLineEnd = true;
			TEXTLINE line = m_page.GetAt(m_currPos.pLine);
			while (!IsLineLastChar(line, m_currPos.charPos))
				m_caretPos.x += GetNextWidth(&m_currPos.charPos);
			m_caretPos.x += GetNextWidth(&m_currPos.charPos);
			break;
		}
	}

	if (!m_currPos.bIsLineEnd) {
		int nCaretHeight = GetTokenInfoHeight(m_currPos.charPos.pTokenInfo);
		if (nCaretHeight != m_nCaretHeight) {
			m_nCaretHeight = nCaretHeight;
			CreateSolidCaret(2, m_nCaretHeight);
			ShowCaret();
		}
	}

	CString str;
	str.Format("%d", m_nCaretHeight);
	GetParentFrame()->SetMessageText(str);

	SetCaretPos(m_caretPos);
		
	CScrollView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CHTMLViewerView::OnSetFocus(CWnd* pOldWnd) 
{
	CScrollView::OnSetFocus(pOldWnd);
	
	//if (!GetDocument()->m_tokens.IsEmpty())
	//{
		CreateSolidCaret(2, m_nCaretHeight);
		SetCaretPos(m_caretPos);
		ShowCaret();
	//}
}

void CHTMLViewerView::OnKillFocus(CWnd* pNewWnd) 
{
	CScrollView::OnKillFocus(pNewWnd);
	
	HideCaret();
	DestroyCaret();
}

void CHTMLViewerView::OnSize(UINT nType, int cx, int cy) 
{
	CScrollView::OnSize(nType, cx, cy);
	
	if (!m_page.IsEmpty()) {
		BreakLines();
		UpdateCaret();
	}
}

CHARPOS CHTMLViewerView::GetNextCharPos(CHARPOS charPos)
{
	const CTokenInfo& ti = GetTokenList().GetAt(charPos.pTokenInfo);
	
	ASSERT (charPos.nIndex + 1 <= ti.m_nLen);
	
	if (charPos.nIndex + 1 == ti.m_nLen) {
		GetTokenList().GetNext(charPos.pTokenInfo);
		charPos.nIndex = 0;
	}
	else
		charPos.nIndex++;

	return charPos;
}

int CHTMLViewerView::GetNextWidth(CHARPOS* pCharPos)
{
	CTokenInfo* pTI = &GetTokenList().GetAt(pCharPos->pTokenInfo);
	int w = pTI->m_pWidths[pCharPos->nIndex++];
	
	if (pCharPos->nIndex >= pTI->m_nLen) {
		pCharPos->nIndex = 0;
		GetTokenList().GetNext(pCharPos->pTokenInfo);
	}
	
	return w;
}

int CHTMLViewerView::GetPrevWidth(CHARPOS* pCharPos)
{
	CTokenInfo* pTI;
	pCharPos->nIndex--;
	
	if (pCharPos->nIndex < 0) {
		if (pCharPos->pTokenInfo == NULL)
			pCharPos->pTokenInfo = GetTokenList().GetTailPosition();
		else
			GetTokenList().GetPrev(pCharPos->pTokenInfo);

		pTI = &GetTokenList().GetAt(pCharPos->pTokenInfo);
		pCharPos->nIndex = pTI->m_nLen - 1;
	}	
	else
		pTI = &GetTokenList().GetAt(pCharPos->pTokenInfo);
	
	return pTI->m_pWidths[pCharPos->nIndex];
}

bool CHTMLViewerView::IsLineFirstChar(const TEXTLINE& line, const CHARPOS& cp)
{
	return (line.first.pTokenInfo == cp.pTokenInfo &&
			line.first.nIndex == cp.nIndex);
}

bool CHTMLViewerView::IsLineLastChar(const TEXTLINE& line, const CHARPOS& cp)
{
	return (line.last.pTokenInfo == cp.pTokenInfo &&
			line.last.nIndex == cp.nIndex);
}

bool CHTMLViewerView::IsLineEnd(const TEXTLINE& line, const CHARPOS& cp)
{
	if (line.last.pTokenInfo == cp.pTokenInfo)
	{
		if (line.last.nIndex >= cp.nIndex)
			return false;
		else
			return true;
	}
	
	POSITION nextTokenInfo = line.last.pTokenInfo;
	m_page.GetNext(nextTokenInfo);

	if (cp.pTokenInfo == nextTokenInfo &&
		cp.nIndex == 0)
		return true;

	return false;
}

void CHTMLViewerView::UpdateCaret()
{	
	POSITION linePos = m_page.GetHeadPosition();
	// This function shouldn't be called with an empty page
	ASSERT(linePos);

	TOKENLIST& tokList = GetTokenList();
	POSITION tokPos;
	TEXTLINE line;
	CHARPOS cp;

	m_caretPos.y = m_rcMargins.top;

	// If m_currPos.charPos is the page last character,
	// processing could be solved in a few lines
	if (m_currPos.charPos.pTokenInfo == NULL)
	{
		do {
			m_currPos.pLine = linePos;
			line = m_page.GetNext(linePos);
			if (linePos != NULL)
				m_caretPos.y += line.nHeight;
			else
			{
				m_caretPos.x = m_rcMargins.left;
				cp.pTokenInfo = line.first.pTokenInfo;
				cp.nIndex = line.first.nIndex;

				while (cp.pTokenInfo != NULL) 
					m_caretPos.x += GetNextWidth(&cp);
				
				SetCaretPos(m_caretPos);
				return;
			}
		} while (true);
	}

	// The first step below its optional. I wrotte it
	// because makes searching faster

	// 1step: find the first line referencing to current token
	do {
		ASSERT(linePos); // it never must be null
		m_currPos.pLine = linePos;
		line = m_page.GetNext(linePos);
		
		// Look between intermediate tokens
		for (tokPos = line.first.pTokenInfo; 
			tokPos != m_currPos.charPos.pTokenInfo &&
			tokPos != line.last.pTokenInfo; 
			tokList.GetNext(tokPos));

		if (tokPos != m_currPos.charPos.pTokenInfo)
			m_caretPos.y += line.nHeight;
		else
			break;

	} while (true);

	// There is a special case that needs start from the previous line
	// If caret was at line end and currCharPos is at the beginning
	// of the founded line, the new caret position should be at
	// the end of the previous line rather than at the beginning
	// of the current line
	if (m_currPos.bIsLineEnd && line.first.nIndex == m_currPos.charPos.nIndex)
	{
		linePos = m_currPos.pLine;
		m_page.GetPrev(m_currPos.pLine);
		line = m_page.GetAt(m_currPos.pLine);
		m_caretPos.y -= line.nHeight;
	}

	// 2step: beginning from the founded line, 
	// look for current char position
	do {
		cp.pTokenInfo = line.first.pTokenInfo;
		cp.nIndex = line.first.nIndex;
		
		m_caretPos.x = m_rcMargins.left;

		while (!IsLineLastChar(line, cp) && cp != m_currPos.charPos) 
			m_caretPos.x += GetNextWidth(&cp);

		if (cp != m_currPos.charPos) // currCharPos is not founded
		{
			// If caret was at the line end and next char is our goal,
			// since we know thar we are at line end also, we need to
			// let the caret here
			if (m_currPos.bIsLineEnd && m_currPos.charPos == GetNextCharPos(cp))
			{
				m_caretPos.x += GetNextWidth(&cp);
				break;
			}
			else
				m_caretPos.y += line.nHeight;
		}
		else
		{
			m_currPos.bIsLineEnd = false;
			break;
		}

		// m_currPos.charPos always must be founded so linePos
		// is not supposed to be null at this point
		ASSERT(linePos);

		m_currPos.pLine = linePos;
		line = m_page.GetNext(linePos);
	} while (true);

	SetCaretPos(m_caretPos);
}

CPoint CHTMLViewerView::GetCharFromPt(const CPoint& pt, RELCHARPOS& rcp)
{
	TEXTLINE line;
	int y = m_rcMargins.top;

	// Find line
	POSITION linePos = m_page.GetHeadPosition();

	do {
		rcp.pLine = linePos;
		line = m_page.GetNext(linePos);

		if (linePos && y + line.nHeight < pt.y)
			y += line.nHeight;
		else
			break;
	} while (true);

	// Find character
	TOKENLIST& list = GetTokenList();
	CHARPOS cp;;
	CTokenInfo ti;
	int x = m_rcMargins.left, dx;

	cp = rcp.charPos = line.first;
	
	while (!IsLineLastChar(line, cp))
	{
		dx = GetNextWidth(&cp);

		if (x + dx >= pt.x)
		{
			if (pt.x - x >= x + dx - pt.x)
			{
				x += dx;
				rcp.charPos = cp;
			}
			break;
		}
		else
		{
			x += dx;
			rcp.charPos = cp;
		}
	}
	return CPoint(x, y);
}

void CHTMLViewerView::MoveCaretTo(const CPoint& pt, const RELCHARPOS& cp)
{
	m_currPos = cp;
	m_currPos.bIsLineEnd = false;

	int nCaretHeight = GetTokenInfoHeight(m_currPos.charPos.pTokenInfo);
	
	if (nCaretHeight != m_nCaretHeight) {
		m_nCaretHeight = nCaretHeight;
		CreateSolidCaret(2, m_nCaretHeight);
		ShowCaret();
	}

	m_caretPos = pt;

	SetCaretPos(m_caretPos);
}

void CHTMLViewerView::OnMouseMove(UINT nFlags, CPoint point) 
{	
	if (!m_page.IsEmpty())
	{
		RELCHARPOS rcp;
		
		GetCharFromPt(point, rcp);

		if (GetTokenList().GetAt(rcp.charPos.pTokenInfo).m_nStyle & TI_HYPERLINK)
			::SetCursor(AfxGetApp()->LoadCursor(IDC_HAND));
		else
			::SetCursor(::LoadCursor(NULL, IDC_ARROW));
	}
	else
		::SetCursor(::LoadCursor(NULL, IDC_ARROW));

	CScrollView::OnMouseMove(nFlags, point);
}

void CHTMLViewerView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	m_selStart = m_currPos;
	
	CPoint pt = GetCharFromPt(point, m_selEnd);

	m_selEnd.pt = pt;

	m_bSelection = true;
	
	
	CScrollView::OnRButtonDown(nFlags, point);
}
